#pragma once

#include<dune/pdelab/common/quadraturerules.hh>
#include<dune/pdelab/finiteelement/localbasiscache.hh>
#include<dune/pdelab/localoperator/defaultimp.hh>
#include<dune/pdelab/localoperator/pattern.hh>
#include<dune/pdelab/localoperator/flags.hh>
#include<dune/pdelab/localoperator/variablefactories.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>



template<typename Param, typename CHGFS, typename CHContainer, typename NSGFS, typename NSContainer>
    class FEM_FF_CHNS_NavierStokes :
      public Dune::PDELab::FullVolumePattern,
      public Dune::PDELab::LocalOperatorDefaultFlags,
      public Dune::PDELab::NumericalJacobianVolume<FEM_FF_CHNS_NavierStokes<Param, CHGFS, CHContainer, NSGFS, NSContainer> >,
      public Dune::PDELab::NumericalJacobianApplyVolume<FEM_FF_CHNS_NavierStokes<Param, CHGFS, CHContainer, NSGFS, NSContainer> >
    {
    private:
      typedef typename CHContainer::field_type RF;

      typedef typename CHGFS::template Child<0>::Type FEM;
      typedef typename FEM::Traits::FiniteElementType::Traits::LocalBasisType LocalBasis;
      Dune::PDELab::LocalBasisCache<LocalBasis> cache;

      typedef typename NSGFS::template Child<0>::Type::template Child<0>::Type VFEM;
      typedef typename VFEM::Traits::FiniteElementType::Traits::LocalBasisType VLocalBasis;
      Dune::PDELab::LocalBasisCache<VLocalBasis> vcache;

      Param& param; // parameter functions

      //FOR EVALUATING the new and old cahn-Hilliard values
      typedef Dune::PDELab::LocalFunctionSpace<CHGFS> CHLFS;
      mutable CHLFS chLfs;
      typedef Dune::PDELab::LFSIndexCache<CHLFS> CHLFSCache;
      mutable CHLFSCache chLfsCache;
      typedef typename CHContainer::template ConstLocalView<CHLFSCache> CHView;
      mutable CHView chView;
      mutable CHView chOldView;
      mutable std::vector<typename LocalBasis::Traits::RangeFieldType> chLocal;
      mutable std::vector<typename LocalBasis::Traits::RangeFieldType> chOldLocal;

      //FOR EVALUATING the old navier-Stokes values
      typedef Dune::PDELab::LocalFunctionSpace<NSGFS> NSLFS;
      mutable NSLFS nsLfs;
      typedef Dune::PDELab::LFSIndexCache<NSLFS> NSLFSCache;
      mutable NSLFSCache nsLfsCache;
      typedef typename NSContainer::template ConstLocalView<NSLFSCache> NSView;
      mutable NSView nsView;
      mutable std::vector<typename VLocalBasis::Traits::RangeFieldType> nsOldLocal;

      //FOR DETECTING ENTITY changes
      typedef typename NSGFS::Traits::GridViewType::template Codim<0>::Entity   ENT;
      mutable ENT OldEnt;

    public:
        // pattern assembly flags
      enum { doPatternVolume = true };

      // residual assembly flags
      enum { doAlphaVolume = true };


      FEM_FF_CHNS_NavierStokes (Param& param_, CHGFS& chGfs_, CHContainer& chV_, CHContainer& chVOld_,
                         NSGFS& nsGfs_, NSContainer& nsVOld_) :
        param(param_),
        chLfs(chGfs_), chLfsCache(chLfs), chView(chV_), chOldView(chVOld_),
        chLocal(chGfs_.maxLocalSize()), chOldLocal(chGfs_.maxLocalSize()),
        nsLfs(nsGfs_), nsLfsCache(nsLfs), nsView(nsVOld_), nsOldLocal(nsGfs_.maxLocalSize())
      {
      }


      template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
      void alpha_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r) const
      {
        //Left hand side, terms depending on the solution x
        using namespace Dune::Indices;

        //Quadrature Rule
        const int dim = EG::Entity::dimension;
        const int order = 2; //TODO: Choose more educated
        auto rule = Dune::PDELab::quadratureRule(eg.geometry(), order);
        //const auto& myordering = lfsv.gridFunctionSpace().ordering();

        const auto& pfspace = child(chLfs,_0);
        const auto& xispace = child(chLfs,_1);
        const auto& femspace = child(chLfs,_0);

        const auto& vspace = child(lfsv,_0);
        const auto& pspace = child(lfsv,_1);
        const auto& vfemspace = child(vspace,_0);

        if(eg.entity() != OldEnt)
        {
          OldEnt = eg.entity();

          chLfs.bind(eg.entity());
          chLfsCache.update();

          chView.bind(chLfsCache);
          chView.read(chLocal);
          chView.unbind();

          chOldView.bind(chLfsCache);
          chOldView.read(chOldLocal);
          chOldView.unbind();

          nsLfs.bind(eg.entity());
          nsLfsCache.update();

          nsView.bind(nsLfsCache);
          nsView.read_sub_container(vspace,nsOldLocal);
          nsView.unbind();
        }

        for(const auto& ip : rule)
        {
          // evaluate shape functions and  gradient of shape functions
          auto& phi = cache.evaluateFunction(ip.position(), femspace.finiteElement().localBasis());
          auto& gradphihat = cache.evaluateJacobian(ip.position(), femspace.finiteElement().localBasis());

          auto& vphi = vcache.evaluateFunction(ip.position(), vfemspace.finiteElement().localBasis());
          auto& gradvphihat = vcache.evaluateJacobian(ip.position(), vfemspace.finiteElement().localBasis());

          // transform gradients of shape functions to real element
          const auto S = eg.geometry().jacobianInverseTransposed(ip.position());
          auto gradphi = makeJacobianContainer(femspace);
          auto gradvphi = makeJacobianContainer(vfemspace);
          for (size_t i=0; i<femspace.size(); i++)
          {
            S.mv(gradphihat[i][0],gradphi[i][0]);
          }
          for (size_t i=0; i<vfemspace.size(); i++)
          {
            S.mv(gradvphihat[i][0],gradvphi[i][0]);
          }

          //compute current values of the unknowns
          Dune::FieldVector<RF,dim> v(0.0);
          Dune::FieldVector<RF,dim> vOld(0.0);
          Dune::FieldMatrix<RF,dim,dim> jacv(0.0);
          RF p=0.0;
          RF pf=0.0;
          Dune::FieldVector<RF,dim> gradpf(0.0);
          RF pfOld=0.0;
          Dune::FieldVector<RF,dim> gradxi(0.0);

          for (size_t i=0; i<femspace.size(); i++)
          {

            p += x(pspace,i)*phi[i];

            pf += chLocal[pfspace.localIndex(i)]*phi[i];
            gradpf.axpy(chLocal[pfspace.localIndex(i)],gradphi[i][0]);
            pfOld += chOldLocal[pfspace.localIndex(i)]*phi[i];
            gradxi.axpy(chLocal[xispace.localIndex(i)],gradphi[i][0]);
          }
          for (size_t i=0; i<vfemspace.size(); i++)
          {
            for(size_t k=0; k<dim; k++)
            {
              v[k] += x(child(vspace,k),i)*vphi[i];
              vOld[k] += nsOldLocal[child(vspace,k).localIndex(i)]*vphi[i];
              jacv[k].axpy(x(child(vspace,k),i),gradvphi[i][0]);
            }
          }

          Dune::FieldVector<RF,dim> g(0.0);

          //Dune::FieldVector<RF, dim> xg = eg.geometry().global( ip.position() );
          //g[1]= 20*xg[0]*(1-xg[0]);
          g[1] = -5;
          RF rho=param.phys.rho_f(pf,1-pf);
          RF rhoOld=param.phys.rho_f(pfOld,1-pfOld);
          RF realPf = pf;
          //pf=0.01+0.99*pf; //!!!!

          auto factor = ip.weight()*eg.geometry().integrationElement(ip.position());
          for (size_t i=0; i<femspace.size(); i++)
          {
            for(int k=0; k<dim; k++)
            {
              // integrate: (div v)*phi
              r.accumulate(pspace,i, -1.0* (v*gradphi[i][0])*factor);
            }
          }
          for (size_t i=0; i<vfemspace.size(); i++)
          {
            for(int k=0; k<dim; k++)
            {
              // integrate: ( (rho+rhoOld)/2 * v_n+1 - rhoOld * v_n)/dt * phi   + nu * Dv_n+1 : D phi - p_n+1 div phi

              //std::cout << (param.nu * (jacv[k]*gradvphi[i][0]) - p * gradvphi[i][0][k]
              //- param.rho*g[k]*vphi[i] )*factor << std::endl;

              //RF lambda = 3;
              const auto v_nabla_v = v * jacv[k];

              r.accumulate(child(vspace,k),i, factor*(
                              ((rho+rhoOld)/2*v[k]-rhoOld*vOld[k])*vphi[i]/param.time.dt          //TimeEvolution
                              + ( rho* v_nabla_v  + param.mobility.eval(realPf)*gradxi[k]*(param.phys.rho_1-param.phys.rho_2)  )* vphi[i] // Momentum transport
                              - p * gradvphi[i][0][k]            //Pressure
                              + param.phys.mu * (jacv[k]*gradvphi[i][0]) //+vphi[i]*(jacv[k]*gradpf)   )  //Viscosity
                              - rho*g[k]*vphi[i]                //Gravitation
                              + param.SurfaceTensionScaling*pfOld*gradxi[k]*vphi[i]    //Surface Tension
                        ));
            }
          }
        }
      }

    };
