
#pragma once
#include <cmath>
#include <dune/pdelab/common/function.hh>
#include <dune/pdelab/constraints/common/constraintsparameters.hh>

#include <dune/phasefield/initial_fn.hh>
#include <dune/phasefield/boundary_rectangular.hh>



template<typename GV, typename RF, typename Params, int dim>
class VelocityInitial_InOutFlow :
  public Dune::PDELab::AnalyticGridFunctionBase<
  Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim>,
  VelocityInitial_InOutFlow<GV,RF,Params,dim> >
{
  const Params_Time& time;
  RF v;
  RF yS;
  RF dSolid;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim> Traits;

  typedef typename Traits::DomainType DomainType;
  typedef typename Traits::RangeType RangeType;

  VelocityInitial_InOutFlow(const GV & gv, const Params & params) :
    Dune::PDELab::AnalyticGridFunctionBase<Traits,VelocityInitial_InOutFlow<GV,RF,Params,dim> > (gv)
    , time(params.time)
  {
    //TODO: Use right scaling (now small coordinates)
    v = params.gradP/params.phys.mu;
    yS = params.YScaling;
    dSolid = params.initial.ptree.get("dSolid",0.3);
  }

  template <typename T>
  void setTime(T t){
    time = t;
  }

  inline void evaluateGlobal(const DomainType & x, RangeType & y) const
  {
    y[0] = std::max(0.0, v/2.0*(x[1]-dSolid*yS)*((2*yS-dSolid*yS)-x[1]));
  }
};

template<typename GV, typename RF, typename Params>
class PInitial_InOutFlow :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PInitial_InOutFlow<GV,RF,Params> >
{
  RF pConst;
public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PInitial_InOutFlow(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PInitial_InOutFlow<GV,RF,Params> > (gv)
     {
       pConst = params.initial.ptree.get("pConst",(RF)0.01);
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y = 0;
    /*if(x[0]<1e-7)
    {
      y=pConst/(yS*yS);
    }*/
  }
};

template<typename GV, typename RF, typename Params, const int Phase>
class PhiInitial_Thinstrip_Layers :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_Thinstrip_Layers<GV,RF,Params,Phase> >
{
  RF eps;
  RF dSolid;
  RF d1;
  const Params & pref;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_Thinstrip_Layers(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_Thinstrip_Layers<GV,RF,Params,Phase> > (gv),
     pref(params)
     {
       eps = params.eps;
       dSolid = params.initial.ptree.get("dSolid",0.3);
       d1 = params.initial.ptree.get("d1",0.4);
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    //RF x_nondim = x[1]/pref.YScaling;
    RF d_sine = 0.15*sin(x[0]*2*M_PI);
    RF rf = pref.tw.dw.eval_shape(1/eps*(x[1] - dSolid*pref.YScaling));
    RF rf2 = pref.tw.dw.eval_shape(1/eps*(x[1] - (dSolid + d1+d_sine)*pref.YScaling));
    if(Phase == 1)
    {
      y = rf*rf2;
      return;
    }
    y=rf*(1-rf2);
  }

  void ApplyReactionDt(const Params & params)
  {
    RF uIn = params.initial.ptree.get("uInflow",(RF)0.5);
    dSolid += params.reaction.eval(uIn)*params.time.dt;
    d1 -= params.reaction.eval(uIn)*params.time.dt;
    std::cout << "New dSolid " << dSolid << " , changed by " << params.reaction.eval(uIn) << " times dt";
  }
};
